#include <iostream>
#include <cassert>
#include <stack>
#include <map>
#include <algorithm>
#include "ast.hpp"
using namespace std;
// Koopa IR % 计数
int cnt = -1;
int ifcnt = 0;
int while_cnt = 0;
int endcnt = 0;
int is_end = 0; // 当前作用域是否结束
struct Symbol
{
    int type;
    int data = 0;
    int level;
    Symbol() {}
    Symbol(int type_, int level_, int data_ = 0)
    {
        type = type_;
        data = data_;
        level = level_;
    }
};

stack<map<string, Symbol>> symbolStack; // 符号表栈
stack<int> while_stack;
map<string, int> name_label;
Symbol SearchSymbol(string s, int value, int ifInsert);
void CompUnitAST::Dump() const
{
    func_def->Dump();
}

void FuncDefAST::Dump() const
{
    cout << "fun ";
    cout << "@" << ident << "(): i32 {" << endl;
    cout << "%entry:" << endl;
    block->Dump();
    if (is_end == 0)
    {
        cout << "  %" << cnt + 1 << " = add 0, 0" << endl;
        cnt++;
        cout << "  ret %" << cnt << endl;
    }
    cout << "}";
}
void FuncTypeAST::Dump() const
{
}
void BlockAST::Dump() const
{

    map<string, Symbol> symbols;
    symbolStack.push(symbols);
    for (auto &item : *blockItems)
    {
        item->Dump();
    }
    symbolStack.pop();
}

void BlockItemAST::Dump() const
{
    if (is_end == 0)
    {
        if (type_ == 1)
            decl->Dump();
        if (type_ == 3)
            ifStmt->Dump();
    }
}

// Stmt        ::= "return" Exp ";";
void StmtAST::Dump() const
{

    if (type_ == 1) // return exp
    {

        exp->Dump();
        cout << "  ret %" << cnt << endl;
        is_end = 1;
    }
    else if (type_ == 2) // lval=exp
    {
        string ident_ = dynamic_cast<LValAST *>(lVal.get())->ident;
        exp->Dump();
        int value_ = dynamic_cast<ExpAST *>(exp.get())->value();
        Symbol symbol = SearchSymbol(ident_, value_, 1); //
        string koopa_ident = ident_ + "_" + to_string(symbol.level);
        cout << "  store %" << cnt << ", @" << koopa_ident << endl;
        // cout << koopa_ident << "=" << value_ << endl;
    }
    else if (type_ == 3) // block
    {
        block->Dump();
    }
    else if (type_ == 4) // exp
    {
        exp->Dump();
    }
    else if (type_ == 6) // return ;
    {
        cout << "  ret " << endl;
        is_end = 1;
    }
    else if (type_ == 7) // WHILE
    {

        int cnt1 = while_cnt;
        cout << "  jump %while_entry" << cnt1 << endl;
        cout << "%while_entry" << cnt1 << ":" << endl;
        is_end = 0;
        exp->Dump();

        while_cnt += 1;
        cout << "  br %" << cnt << ", %while_body" << cnt1 << ", %while_end" << cnt1 << endl;
        cout << "%while_body" << cnt1 << ":" << endl;
        is_end = 0;
        while_stack.push(while_cnt - 1);
        stmt->Dump();
        while_stack.pop();
        if (is_end == 0) // 没有结束,回到入口
        {
            cout << "  jump %while_entry" << cnt1 << endl;
        }
        else
        {
            is_end = 0;
        }
        cout << "%while_end" << cnt1 << ":" << endl;
    }
    else if (type_ == 8) // break;
    {
        is_end = 1;
        cout << "  jump %while_end" << while_stack.top() << endl;
    }
    else if (type_ == 9) // continue;
    {
        is_end = 1;
        cout << "  jump %while_entry" << while_stack.top() << endl;
    }
}

void IfStmtAST::Dump() const
{
    if (type_ == 1) // matchedIf
    {
        matchedIf->Dump();
    }
    else if (type_ == 2) // OpenIf
    {
        openIf->Dump();
    }
}

void MatchedIfAST::Dump() const
{
    if (type_ == 1) // stmt
    {
        // cout << 1 << endl;
        stmt->Dump();
    }
    else if (type_ == 2) // IF '(' Exp ')' MatchedIf ELSE MatchedIf
    {
        // cout << 2 << endl;
        exp->Dump();
        int ifcnt1 = ifcnt;

        ifcnt += 1;
        cout << "  br %" << cnt << ", %then" << ifcnt1 << ", %else" << ifcnt1 << endl;
        cout << "%then" << ifcnt1 << ":" << endl;
        is_end = 0;
        matchedIf1->Dump();

        int temp = 0;
        if (is_end == 0) // 没有return
        {
            cout << "  jump %end" << ifcnt1 << endl;
        }
        else
        {
            is_end = 0;
            temp++;
        }

        cout << "%else" << ifcnt1 << ":" << endl;
        is_end = 0;
        matchedIf2->Dump();
        if (is_end == 0) // 没有return
        {
            cout << "  jump %end" << ifcnt1 << endl;
        }
        else
        {
            is_end = 0;
            temp++;
        }

        if (temp < 2) // 两句都结束了就不用输出这个
        {
            cout << "%end" << ifcnt1 << ":" << endl;
            is_end = 0;
        }
        else
        { // 否则整个块都结束
            is_end = 1;
        }
    }
}
void OpenIfAST::Dump() const
{
    if (type_ == 1) // IF '(' Exp ')' IfStmt
    {
        // cout << 3 << endl;
        exp->Dump();
        int ifcnt1 = ifcnt;
        ifcnt += 1;
        cout << "  br %" << cnt << ", %then" << ifcnt1 << ", %end" << ifcnt1 << endl;
        cout << "%then" << ifcnt1 << ":" << endl;
        is_end = 0;
        ifStmt->Dump();
        if (is_end == 0) // 没有return
        {
            cout << "  jump %end" << ifcnt1 << endl;
        }
        cout << "%end" << ifcnt1 << ":" << endl;
        is_end = 0;
    }
    else if (type_ == 2) // IF '(' Exp ')' MatchedIf ELSE OpenIf
    {
        // cout << 4 << endl;
        exp->Dump();
        int ifcnt1 = ifcnt;
        cout << "  br %" << cnt << ", %then" << ifcnt1 << ", %else" << ifcnt1 << endl;
        ifcnt += 1;

        int temp = 0;
        cout << "%then" << ifcnt1 << ":" << endl;
        is_end = 0;
        matchedIf->Dump();
        if (is_end == 0) // 没有return
        {
            cout << "  jump %end" << ifcnt1 << endl;
        }
        else
        {
            is_end = 0;
            temp++;
        }

        cout << "%else" << ifcnt1 << ":" << endl;
        is_end = 0;
        openIf->Dump();
        if (is_end == 0) // 没有return
        {
            cout << "  jump %end" << ifcnt1 << endl;
        }
        else
        {
            is_end = 0;
            temp++;
        }

        if (temp < 2) // 两句都return就不用
        {
            cout << "%end" << ifcnt1 << ":" << endl;
            is_end = 0;
        }
        else
        {
            is_end = 1;
        }
    }
}

// Exp         ::= LOrExp;

// PrimaryExp  ::="(" Exp ")" |  Number | LVal;
void PrimaryExpAST::Dump() const
{
    if (type_ == 1)
    {
        exp->Dump();
    }
    else if (type_ == 2) // 数字，移到%0
    {
        cout << "  %" << cnt + 1 << " = add 0, " << number << endl;
        cnt++;
    }
    else if (type_ == 3)
    {
        lVal->Dump();
    }
}
int PrimaryExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<ExpAST *>(exp.get())->value();
    }
    else if (type_ == 2)
    {
        return number;
    }
    else if (type_ == 3)
    {
        Symbol symbol = SearchSymbol(dynamic_cast<LValAST *>(lVal.get())->ident, 0, 0);
        return symbol.data;
    }
}

// UnaryExp    ::= PrimaryExp | UnaryOp UnaryExp;
// UnaryOp     ::= "+" | "-" | "!";
void UnaryExpAST::Dump() const
{
    if (type_ == 1)
    {
        primaryExp->Dump();
    }
    if (type_ == 2)
    {
        unaryExp->Dump();
        if (unaryOp == '-')
        {
            //%1 = sub 0, %0
            cout << "  %" << cnt + 1 << " = sub 0, %";
            cout << cnt << endl;
            cnt++;
        }
        else if (unaryOp == '!') //+号不用操作
        {
            // %1 = eq 0, %0
            cout << "  %" << cnt + 1 << " = eq 0, %";
            cout << cnt << endl;
            cnt++;
        }
    }
}
int UnaryExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<PrimaryExpAST *>(primaryExp.get())->value();
    }
    else
    {
        int temp = dynamic_cast<UnaryExpAST *>(unaryExp.get())->value();
        if (unaryOp == '-')
        {
            return -temp;
        }
        else if (unaryOp == '!')
        {
            return !temp;
        }
        else if (unaryOp == '+')
        {
            return temp;
        }
    }
}

// MulExp      ::= UnaryExp | MulExp ("*" | "/" | "%") UnaryExp;
void MulExpAST::Dump() const
{
    if (type_ == 1)
    {
        unaryExp->Dump();
    }
    if (type_ == 2)
    {
        unaryExp->Dump();
        int right = cnt;
        mulExp->Dump();
        if (mulOp == '*')
        {
            cout << "  %" << cnt + 1 << " = mul %" << cnt << ", %" << right << endl;
        }
        else if (mulOp == '%')
        {
            cout << "  %" << cnt + 1 << " = mod %" << cnt << ", %" << right << endl;
        }
        else if (mulOp == '/')
        {
            cout << "  %" << cnt + 1 << " = div %" << cnt << ", %" << right << endl;
        }
        cnt++;
    }
}

int MulExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<UnaryExpAST *>(unaryExp.get())->value();
    }
    else
    {
        int left = dynamic_cast<MulExpAST *>(mulExp.get())->value();
        int right = dynamic_cast<UnaryExpAST *>(unaryExp.get())->value();
        if (mulOp == '*')
            return left * right;
        else if (mulOp == '/')
            return left / right;
        else if (mulOp == '%')
            return left % right;
    }
}
// AddExp      ::= MulExp | AddExp ("+" | "-") MulExp;
void AddExpAST::Dump() const
{
    if (type_ == 1)
    {
        mulExp->Dump();
    }
    if (type_ == 2)
    {
        mulExp->Dump();
        int right = cnt;
        addExp->Dump();
        if (addOp == '+')
        {
            cout << "  %" << cnt + 1 << " = add %" << cnt << ", %" << right << endl;
            cnt++;
        }
        else if (addOp == '-')
        {
            cout << "  %" << cnt + 1 << " = sub %" << cnt << ", %" << right << endl;
            cnt++;
        }
    }
}
int AddExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<MulExpAST *>(mulExp.get())->value();
    }
    else
    {
        int left = dynamic_cast<AddExpAST *>(addExp.get())->value();
        int right = dynamic_cast<MulExpAST *>(mulExp.get())->value();
        if (addOp == '+')
            return left + right;
        else if (addOp == '-')
            return left - right;
    }
}

void RelExpAST::Dump() const

{
    if (type_ == 1)
    {
        addExp->Dump();
    }
    if (type_ == 2)
    {
        relExp->Dump();
        int temp = cnt;
        addExp->Dump();
        if (relOp == ">")
        {
            cout << "  %" << cnt + 1 << " = gt %" << temp << ", %" << cnt << endl;
        }
        else if (relOp == "<")
        {
            cout << "  %" << cnt + 1 << " = lt %" << temp << ", %" << cnt << endl;
        }
        else if (relOp == ">=")
        {
            cout << "  %" << cnt + 1 << " = ge %" << temp << ", %" << cnt << endl;
        }
        else if (relOp == "<=")
        {
            cout << "  %" << cnt + 1 << " = le %" << temp << ", %" << cnt << endl;
        }
        cnt++;
    }
}
int RelExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<AddExpAST *>(addExp.get())->value();
    }
    else
    {
        int left = dynamic_cast<RelExpAST *>(relExp.get())->value();
        int right = dynamic_cast<AddExpAST *>(addExp.get())->value();
        if (relOp == ">")
        {
            return left > right;
        }
        else if (relOp == "<")
        {
            return left < right;
        }
        else if (relOp == ">=")
        {
            return left >= right;
        }
        else if (relOp == "<=")
        {
            return left <= right;
        }
    }
}

void EqExpAST::Dump() const
{
    if (type_ == 1)
    {
        relExp->Dump();
    }
    if (type_ == 2)
    {
        eqExp->Dump();
        int temp = cnt;
        relExp->Dump();
        if (eqOp == "==")
        {
            cout << "  %" << cnt + 1 << " = eq %" << cnt << ", %" << temp << endl;
        }
        else if (eqOp == "!=")
        {
            cout << "  %" << cnt + 1 << " = ne %" << cnt << ", %" << temp << endl;
        }
        cnt++;
    }
}
int EqExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<RelExpAST *>(relExp.get())->value();
    }
    else
    {
        int temp1 = dynamic_cast<EqExpAST *>(eqExp.get())->value();
        int temp2 = dynamic_cast<RelExpAST *>(relExp.get())->value();
        if (eqOp == "==")
        {
            return temp1 == temp2;
        }
        else if (eqOp == "!=")
        {
            return temp1 != temp2;
        }
    }
}
void LAndExpAST::Dump() const
{
    if (type_ == 1)
    {
        eqExp->Dump();
    }
    if (type_ == 2)
    {
        //(A!=0)|(B!=0)
        landExp->Dump();
        int temp = dynamic_cast<LAndExpAST *>(landExp.get())->value();
        int left = cnt;
        eqExp->Dump();
        int right = cnt;
        cout << "  %" << cnt + 1 << " = ne %" << left << ", " << 0 << endl;
        cnt++;
        if (temp != 0) //
        {
            cout << "  %" << cnt + 1 << " = ne %" << right << ", " << 0 << endl; // 短路：有待补充
            cnt++;
            cout << "  %" << cnt + 1 << " = and %" << cnt << ", %" << cnt - 1 << endl;
            cnt++;
        }
    }
}
int LAndExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<EqExpAST *>(eqExp.get())->value();
    }
    else
    {
        int temp1 = dynamic_cast<LAndExpAST *>(landExp.get())->value();
        int temp2 = dynamic_cast<EqExpAST *>(eqExp.get())->value();
        return temp1 && temp2;
    }
}
void LOrExpAST::Dump() const
{
    if (type_ == 1)
    {
        landExp->Dump();
    }
    if (type_ == 2)
    {
        //(A!=0)|(B!=0)

        lorExp->Dump();
        int temp = dynamic_cast<LOrExpAST *>(lorExp.get())->value();
        int left = cnt;
        landExp->Dump();
        int right = cnt;
        cout << "  %" << cnt + 1 << " = ne %" << left << ", " << 0 << endl;
        cnt++;
        if (temp == 0) // 只有第一项为0时需要计算第二项，否则就短路
        {
            cout << "  %" << cnt + 1 << " = ne %" << right << ", " << 0 << endl; // 短路：有待补充
            cnt++;
            cout << "  %" << cnt + 1 << " = or %" << cnt << ", %" << cnt - 1 << endl;
            cnt++;
        }
    }
}
int LOrExpAST::value()
{
    if (type_ == 1)
    {
        return dynamic_cast<LAndExpAST *>(landExp.get())->value();
    }
    else
    {
        int temp1 = dynamic_cast<LOrExpAST *>(lorExp.get())->value();
        int temp2 = dynamic_cast<LAndExpAST *>(landExp.get())->value();
        return temp1 || temp2;
    }
}
void ExpAST::Dump() const
{
    lorExp->Dump();
}
int ExpAST::value()
{
    return dynamic_cast<LOrExpAST *>(lorExp.get())->value();
}

void ConstExpAST::Dump() const
{
    exp->Dump();
}
int ConstExpAST::value()
{
    return dynamic_cast<ExpAST *>(exp.get())->value();
}

void ConstInitValAST::Dump() const
{
    constExp->Dump();
}

int ConstInitValAST::value()
{
    return dynamic_cast<ConstExpAST *>(constExp.get())->value();
}

void ConstDefAST::Dump() const
{
    int temp = dynamic_cast<ConstInitValAST *>(constInitVal.get())->value();
    int level = name_label[ident]++;
    Symbol symbol(1, level, temp);
    map<string, Symbol> symbols = symbolStack.top();
    symbols.insert({ident, symbol});
    symbolStack.pop();
    symbolStack.push(symbols);
}
void ConstDeclAST::Dump() const
{
    for (auto &item : *constDefs)
    {
        item->Dump();
    }
}
void DeclAST::Dump() const
{
    if (type_ == 1)
        constDecl->Dump();
    else if (type_ == 2)
        varDecl->Dump();
}
void VarDeclAST::Dump() const
{
    for (auto &item : *varDefs)
    {
        item->Dump();
    }
}
void VarDefAST::Dump() const
{
    int label = name_label[ident]++;
    string koopa_ident = ident + "_" + to_string(label);
    if (type_ == 2) // 有初始值
    {
        int value_ = dynamic_cast<InitValAST *>(initVal.get())->value();

        cout << "  @" << koopa_ident << " = alloc i32 " << endl;
        cout << "  store " << value_ << ", @" << koopa_ident << endl;

        Symbol symbol(2, label, value_);
        map<string, Symbol> symbols = symbolStack.top();
        symbols.insert({ident, symbol});
        symbolStack.pop();
        symbolStack.push(symbols);
    }
    else
    {

        cout << "  @" << koopa_ident << " = alloc i32 " << endl;

        Symbol symbol(2, label);
        map<string, Symbol> symbols = symbolStack.top();
        symbols.insert({ident, symbol});
        symbolStack.pop();
        symbolStack.push(symbols);
    }
}
void InitValAST::Dump() const
{
    exp->Dump();
}

int InitValAST::value()
{
    return dynamic_cast<ExpAST *>(exp.get())->value();
}
void LValAST::Dump() const
{

    Symbol it = SearchSymbol(ident, 0, 0);
    int type_ = it.type, data = it.data;
    if (type_ == 1)
    {
        cout << "  %" << cnt + 1 << " = add 0, " << data << endl;
        cnt++;
    }
    else if (type_ == 2)
    {
        string koopa_ident = ident + "_" + to_string(it.level);
        cout << "  %" << cnt + 1 << " = load @" << koopa_ident << endl;
        cnt++;
    }
}

Symbol SearchSymbol(string s, int value, int ifInsert)
{

    stack<map<string, Symbol>> tempStack; // 临时栈
    Symbol temp;
    map<string, Symbol> tempmap;
    bool flag = false;
    while (!symbolStack.empty())
    {
        tempmap = symbolStack.top();
        map<string, Symbol>::iterator it = tempmap.find(s);
        if (it != tempmap.end()) // 找到了
        {

            temp = it->second;
            // cout << "in func:" << s << "=" << temp.data << endl;
            flag = true;
            if (ifInsert > 0) // 查询的时候修改
            {
                // cout << "in func:" << s << "change into" << value << endl;
                it->second.data = value;
            }
            map<string, Symbol>::iterator it1 = tempmap.find(s);
            // cout << "in func:" << s << " it =" << it1->second.data << endl;
            tempStack.push(tempmap);
            symbolStack.pop();
            break;
        }

        tempStack.push(tempmap);
        symbolStack.pop();
    }
    if (!flag)
    {
        temp.type = -1;
    }

    while (!tempStack.empty())
    {
        map<string, Symbol> tempmap = tempStack.top();
        symbolStack.push(tempmap);
        tempStack.pop();
    }

    return temp;
}