#pragma once
// 所有 AST 的基类
#include <iostream>
#include <memory>
#include <string>
#include <vector>
using namespace std;
class BaseAST
{
public:
    virtual ~BaseAST() = default;
    virtual void Dump() const = 0;
    virtual int value() {}
};

// CompUnit 是 BaseAST
class CompUnitAST : public BaseAST
{
public:
    // 用智能指针管理对象
    unique_ptr<BaseAST> func_def;
    void Dump() const override;
};

// FuncDef 也是 BaseAST
class FuncDefAST : public BaseAST
{
public:
    unique_ptr<BaseAST> func_type;
    string ident;
    unique_ptr<BaseAST> block;
    void Dump() const override;
};

class FuncTypeAST : public BaseAST
{
public:
    string type_;
    void Dump() const override;
};

class BlockAST : public BaseAST
{
public:
    unique_ptr<vector<unique_ptr<BaseAST>>> blockItems;
    void Dump() const override;
};

class BlockItemAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> decl; // type 1
    unique_ptr<BaseAST> stmt;
    unique_ptr<BaseAST> ifStmt;
    void Dump() const override;
};
/*
Stmt ::= LVal "=" Exp ";"
       | [Exp] ";"
       | Block
       | "return" [Exp] ";";
*/

class StmtAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> lVal;
    unique_ptr<BaseAST> exp;
    unique_ptr<BaseAST> block;
    unique_ptr<BaseAST> stmt;
    void Dump() const override;
};
class IfStmtAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> matchedIf;
    unique_ptr<BaseAST> openIf;
    void Dump() const override;
};

class MatchedIfAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> exp;
    unique_ptr<BaseAST> matchedIf1;
    unique_ptr<BaseAST> matchedIf2;
    unique_ptr<BaseAST> stmt;
    void Dump() const override;
};

class OpenIfAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> exp;
    unique_ptr<BaseAST> ifStmt;
    unique_ptr<BaseAST> openIf;
    unique_ptr<BaseAST> matchedIf;
    void Dump() const override;
};

class LOrExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> landExp;
    unique_ptr<BaseAST> lorExp;
    void Dump() const override;
    int value() override;
};

class LAndExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> landExp;
    unique_ptr<BaseAST> eqExp;
    void Dump() const override;
    int value() override;
};

class EqExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> relExp;
    unique_ptr<BaseAST> eqExp;
    string eqOp;
    void Dump() const override;
    int value() override;
};

class RelExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> addExp;
    unique_ptr<BaseAST> relExp;
    string relOp;
    void Dump() const override;
    int value() override;
};

class AddExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> addExp;
    unique_ptr<BaseAST> mulExp;
    char addOp;
    void Dump() const override;
    int value() override;
};

class MulExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> unaryExp;
    unique_ptr<BaseAST> mulExp;
    char mulOp;
    void Dump() const override;
    int value() override;
};

class UnaryExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> primaryExp;
    char unaryOp;
    unique_ptr<BaseAST> unaryExp;
    void Dump() const override;
    int value() override;
};

class PrimaryExpAST : public BaseAST
{
public:
    int type_;
    int number;
    unique_ptr<BaseAST> lVal;
    unique_ptr<BaseAST> exp;
    void Dump() const override;
    int value() override;
};

class DeclAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> constDecl;
    unique_ptr<BaseAST> varDecl;
    void Dump() const override;
};

class ConstDeclAST : public BaseAST
{
public:
    string btype;
    unique_ptr<vector<unique_ptr<BaseAST>>> constDefs;
    void Dump() const override;
};

class ConstDefAST : public BaseAST
{
public:
    string ident;
    unique_ptr<BaseAST> constInitVal;
    void Dump() const override;
};

class VarDeclAST : public BaseAST
{
public:
    string btype;
    unique_ptr<vector<unique_ptr<BaseAST>>> varDefs;
    void Dump() const override;
};

class VarDefAST : public BaseAST
{
public:
    int type_;
    string ident;
    unique_ptr<BaseAST> initVal;
    void Dump() const override;
};

class InitValAST : public BaseAST
{
public:
    unique_ptr<BaseAST> exp;
    void Dump() const override;
    int value() override;
};

class ExpAST : public BaseAST
{
public:
    int type_;
    unique_ptr<BaseAST> lorExp;
    void Dump() const override;
    int value() override;
};

class ConstExpAST : public BaseAST
{
public:
    unique_ptr<BaseAST> exp;
    void Dump() const override;
    int value() override;
};

class ConstInitValAST : public BaseAST
{
public:
    unique_ptr<BaseAST> constExp;
    void Dump() const override;
    int value() override;
};

class LValAST : public BaseAST
{
public:
    string ident;
    void Dump() const override;
};
// ...
