#include <cassert>
#include <iostream>
#include <vector>
#include <unordered_map>
#include <cstring>
#include <algorithm>
#include <map>
#include "raw.hpp"
#include "koopa.h"
using namespace std;
static unordered_map<koopa_raw_value_t, int> location_; // 存放变量地址
static int stackaddr = 0;

// 将value的地址存进reg
void get_addr_to_reg(const koopa_raw_value_t &value, const string &reg)
{
    int imm = location_[value];
    if (imm < 2048 && imm >= -2047)
        cout << "  addi " << reg << ", sp, " << imm << endl; // sp+imm
    else
    {
        cout << "  li t2, " << imm << endl;
        cout << "  add " << reg << " , sp , t2 " << endl;
    }
}

// 存储在 reg的value值进入内存
void save_to_memory(const koopa_raw_value_t &value, const string &reg)
{

    get_addr_to_reg(value, "t6");
    cout << "  sw " << reg << ", 0(t6)" << endl;
}

// value 转移到寄存器
void load_to_reg(const koopa_raw_value_t &value, const string &reg)
{
    if (value->kind.tag == KOOPA_RVT_INTEGER) // 立即数：直接移到寄存器
    {
        cout << "  li " << reg << ", " << value->kind.data.integer.value << endl;
    }
    else // 在栈中
    {
        get_addr_to_reg(value, "t6");
        cout << "  lw " << reg << ", 0(t6)" << std::endl;
    }
}

void Visit(const koopa_raw_program_t &program)
{

    // 执行一些其他的必要操作
    // ...
    // 访问所有全局变量
    Visit(program.values); // type: koopa_raw_slice_t
    // 访问所有函数
    Visit(program.funcs);
}
// 访问 raw slice
void Visit(const koopa_raw_slice_t &slice)
{

    for (size_t i = 0; i < slice.len; ++i)
    {
        auto ptr = slice.buffer[i];
        // 根据 slice 的 kind 决定将 ptr 视作何种元素
        switch (slice.kind)
        {
        case KOOPA_RSIK_FUNCTION:
            // 访问函数
            Visit(reinterpret_cast<koopa_raw_function_t>(ptr));
            break;
        case KOOPA_RSIK_BASIC_BLOCK:
            // 访问基本块
            Visit(reinterpret_cast<koopa_raw_basic_block_t>(ptr));
            break;
        case KOOPA_RSIK_VALUE:
            // 访问指令
            Visit(reinterpret_cast<koopa_raw_value_t>(ptr));
            break;

        default:
            // 我们暂时不会遇到其他内容, 于是不对其做任何处理
            assert(false);
        }
    }
}
// 访问函数
void Visit(const koopa_raw_function_t &func)
{

    cout << "  .text" << endl;
    cout << "  .globl " << func->name + 1 << endl;
    cout << func->name + 1 << ":" << endl;
    // 执行一些其他的必要操作
    // ...
    // 访问所有基本块
    Visit(func->bbs); // type of bbs:slice_t stmt?
}

// 访问基本块
void Visit(const koopa_raw_basic_block_t &bb)
{

    // 执行一些其他的必要操作
    // ...
    // 访问所有指令
    cout << bb->name + 1 << ":" << endl;
    Visit(bb->insts); // slice
}

// 访问指令
void Visit(const koopa_raw_value_t &value)
{

    // 根据指令类型判断后续需要如何访问
    /*
    koopa_raw_integer_t integer;
    koopa_raw_aggregate_t aggregate;
    koopa_raw_func_arg_ref_t func_arg_ref;
    koopa_raw_block_arg_ref_t block_arg_ref;
    koopa_raw_global_alloc_t global_alloc;
    koopa_raw_load_t load;
    koopa_raw_store_t store;
    koopa_raw_get_ptr_t get_ptr;
    koopa_raw_get_elem_ptr_t get_elem_ptr;
    koopa_raw_binary_t binary;
    koopa_raw_branch_t branch;
    koopa_raw_jump_t jump;
    koopa_raw_call_t call;
    koopa_raw_return_t ret;
    */

    const auto &kind = value->kind;
    switch (kind.tag)
    {
    case KOOPA_RVT_RETURN:
        // 访问 return 指令
        Visit(kind.data.ret); // koopa_raw_return_t
        break;
    case KOOPA_RVT_INTEGER:
        // 访问 integer 指令
        Visit(kind.data.integer); // koopa_raw_integer_t
        break;
    case KOOPA_RVT_BINARY:
        // 访问binary指令
        Visit(kind.data.binary, value);
        break;
    case KOOPA_RVT_LOAD:
        // 访问load指令
        Visit(kind.data.load, value);
        break;
    case KOOPA_RVT_STORE:
        // 访问store指令
        Visit(kind.data.store);
        break;
    case KOOPA_RVT_ALLOC:
        // 访问 alloc 指令
        Visit("alloc", value);
        break;
    case KOOPA_RVT_JUMP:
        // 访问jump指令
        Visit(kind.data.jump);
        break;
    case KOOPA_RVT_BRANCH:
        // 访问branch指令
        Visit(kind.data.branch, value);
        break;
    default:
        // 其他类型暂时遇不到
        assert(false);
    }
}

// ret指令
void Visit(const koopa_raw_return_t &ret)
{

    load_to_reg(ret.value, "a0");
    cout << "  ret" << endl;
}

// integer
void Visit(const koopa_raw_integer_t &integer)
{
}

// binary
void Visit(const koopa_raw_binary_t &binary, const koopa_raw_value_t &value)
{
    /*
    typedef struct {
      /// Operator.
      koopa_raw_binary_op_t op;
      /// Left-hand side value.
      koopa_raw_value_t lhs;
      /// Right-hand side value.
      koopa_raw_value_t rhs;
    } koopa_raw_binary_t;
    */

    load_to_reg(binary.lhs, "t0");
    load_to_reg(binary.rhs, "t1");
    if (binary.op == KOOPA_RBO_EQ)
    {
        cout << "  xor t0, t0, t1" << endl;
        cout << "  seqz t0, t0" << endl;
    }
    if (binary.op == KOOPA_RBO_NOT_EQ)
    {
        cout << "  xor t0, t0, t1" << endl;
        cout << "  snez t0, t0" << endl;
    }
    if (binary.op == KOOPA_RBO_GT)
    {
        cout << "  sgt t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_LT)
    {
        cout << "  slt t0, t0, t1" << endl;
    }

    if (binary.op == KOOPA_RBO_GE) //>= 不小于
    {
        cout << "  slt t0, t0, t1" << endl;
        cout << "  xori t0, t0, 1" << endl;
    }
    if (binary.op == KOOPA_RBO_LE) // 不大于
    {
        cout << "  sgt t0, t0, t1" << endl;
        cout << "  xori t0, t0, 1" << endl;
    }
    if (binary.op == KOOPA_RBO_ADD)
    {
        cout << "  add t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_SUB)
    {
        cout << "  sub t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_MUL)
    {
        cout << "  mul t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_DIV)
    {
        cout << "  div t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_MOD)
    {
        cout << "  rem t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_AND)
    {
        cout << "  and t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_OR)
    {
        cout << "  or t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_XOR)
    {
        cout << "  xor t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_SHL)
    {
        cout << "  sll t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_SHR)
    {
        cout << "  srl t0, t0, t1" << endl;
    }
    if (binary.op == KOOPA_RBO_SAR)
    {
        cout << "  sra t0, t0, t1" << endl;
    }
    if (value->ty->tag != KOOPA_RTT_UNIT)
    {
        location_[value] = stackaddr;
        stackaddr += 4;
        save_to_memory(value, "t0");
    }
}

//%5 = load @y load指令
void Visit(const koopa_raw_load_t &load, const koopa_raw_value_t &value)
{ /*
 typedef struct {
   /// Source.
   koopa_raw_value_t src; 这里的y
 } koopa_raw_load_t;
 */
    load_to_reg(load.src, "t0");

    // 存入栈
    if (value->ty->tag != KOOPA_RTT_UNIT)
    {
        location_[value] = stackaddr;
        stackaddr += 4;
        save_to_memory(value, "t0");
    }
}
// jump j else
void Visit(const koopa_raw_jump_t &jump)
{
    cout << "  j " << jump.target->name + 1 << endl;
}
// branch  bnez t0, then
void Visit(const koopa_raw_branch_t &branch, const koopa_raw_value_t &value)
{
    cout << "  bnez t0," << branch.true_bb->name + 1 << endl;
    cout << "  j " << branch.false_bb->name + 1 << endl;
}
// store指令
void Visit(const koopa_raw_store_t &store)
{

    load_to_reg(store.value, "t0");    // value存到t0
    get_addr_to_reg(store.dest, "t6"); // 地址存到t6
    std::cout << "  sw t0, 0(t6)" << std::endl;
}

// alloc指令
void Visit(const std::string &alloc, const koopa_raw_value_t &value)
{
    location_[value] = stackaddr;
    stackaddr += 4;
}
